@@ -1,4 +1,5 @@
 """Base validation."""
+import datetime
 import os
 
 VALID_BED_EXTENSIONS = (
@@ -29,8 +30,11 @@ def validate_bam_file(fname, check_exist=False):
         raise ValueError(f"Bam file {fname} does not exist.")
 
 
-def validate_string(value, choices=None):
+def validate_string(value, choices=None, allow_empty=False):
     """Validate string."""
+    if not value and allow_empty:
+        return
+
     if not isinstance(value, str):
         raise ValueError(f"Value {value} should be a string.")
 
@@ -43,3 +47,14 @@ def validate_integer(value):
     """Validate integer."""
     if not isinstance(value, int):
         raise ValueError(f"Value {value} should be an integer.")
+
+
+def validate_date(value, allow_empty=False):
+    """Validate date format."""
+    if not value and allow_empty:
+        return
+
+    try:
+        datetime.datetime.strptime(value, "%Y-%m-%d")
+    except ValueError:
+        raise ValueError(f"Incorrect date format ({value}), should be YYYY-MM-DD.")
